module PetEd

require 'gtk2'
require 'cairo'
require 'pango'
require 'optparse'
require_relative 'pet'
require_relative 'pet_canvas'
require_relative 'pet_attr'
require_relative 'pet_conf_ed'
require_relative 'pet_log'
require_relative 'pet_def'

Mode_Auto = 0
	Mode_NET = 1

# add close button and check button for viewing attribute list
class PetedNotebookWidget < Gtk::Notebook
	def initialize(toolbar_top)
	  super()
		@toolbar_top = toolbar_top
	end

	def append_page(child, name)
		tab = Gtk::HBox.new(false, 0) # (homogeneous, spacing)
		l = Gtk::Label.new(File.basename(name))
		l.tooltip_text = name
		c = Gtk::CheckButton.new 
		c.focus_on_click = false
		c.tooltip_text = 'Attribute View'
		#c.signal_connect('toggled') {|w| child.show_attributes(w.active?)}
		#c.set_size_request(18, 18) # fine but arbitrary
		#i = Gtk::Image.new(Gtk::Stock::CLOSE, Gtk::IconSize::MENU)
		r = Gtk::RcStyle.new
		r.xthickness = 0
		r.ythickness = 0
		b = Gtk::Button.new
		b.tooltip_text = 'Close'
		b.signal_connect('clicked') {
			i = self.page_num(child)
			active = (i == self.page)
			self.remove_page(i)
			if active
				@toolbar_top.switch( self.get_nth_page(self.page))
			end
		}
		b.modify_style(r)
		b.focus_on_click = false
		b.relief = Gtk::RELIEF_NONE
		b.add(Gtk::Image.new(Gtk::Stock::CLOSE, Gtk::IconSize::MENU))
		w, h = Gtk::IconSize.lookup_for_settings(b.settings, Gtk::IconSize::MENU) # as used in gedit-close-button.c (2.30.4)
		b.set_size_request(w + 2, h + 2)
		tab.pack_start(l, false, false, 0) # (child, expand = true, fill = true, padding = 0)
		tab.pack_start(c, false, false, 0)
		tab.pack_start(b, false, false, 0)
		tab.show_all
		super(child, tab)
	end

	def update_label(child, name)
		self.get_tab_label(child).children.each{|w|
			if w.class == Gtk::Label
				w.text = File.basename(name)
				w.tooltip_text = name
				break
			end
		}
	end

	def get_labeltext(child)
	  text = ''
		self.get_tab_label(child).children.each{|w|
			if w.class == Gtk::Label
				text = w.text
				break
			end
		}
		text
	end
end

MenuWindowName = '<PetedMainWindow>'
class PetedMenuItemStock < Gtk::ImageMenuItem
	def initialize(menu, item, id)
		super(id)
		self.accel_path = MenuWindowName + '/' + menu + '/' + item
		modifier, keyval = Gtk::Stock.lookup(id)[2, 2]
		Gtk::AccelMap.add_entry(self.accel_path, keyval, modifier)
	end
end

class PetedMenuItemText < Gtk::ImageMenuItem
	def initialize(menu, itemtext, keyval = nil, modifier = nil)
		super(itemtext)
		self.accel_path = MenuWindowName + '/' + menu + '/' + itemtext
		if keyval and modifier
		  Gtk::AccelMap.add_entry(self.accel_path, keyval, modifier)
		end
	end
end

class PetedMenuBar < Gtk::MenuBar
	def initialize(group, main_window, menu_hint_status_bar)
		super()
		@bar = menu_hint_status_bar
		@id = menu_hint_status_bar.get_context_id('Peted_Menu_Hint')
		title = 'File'
		filemenu = Gtk::Menu.new
		filemenu.set_accel_group(group)
		filemenu.set_accel_path(MenuWindowName + '/' + title + '/')
		file = Gtk::MenuItem.new(title)
		file.submenu = filemenu
		self.append(file)

		name = 'New'
		i = PetedMenuItemStock.new(title, name, Gtk::Stock::NEW)
		i.signal_connect('activate') {|w| main_window.open_schematics([''])}
		set_menu_hint(i, 'Create a new sheet')
		filemenu.append(i)

		name = 'Open...'
		i = PetedMenuItemStock.new(title, name, Gtk::Stock::OPEN)
		i.signal_connect('activate') {|w| main_window.open_schematics(nil)}
		set_menu_hint(i, 'Open a file')
		filemenu.append(i)

		name = 'Add Sym...'
		i = PetedMenuItemText.new(title, name, Gdk::Keyval::GDK_a, Gdk::Window::SHIFT_MASK)
		i.signal_connect('activate') {|w| main_window.add_symbol(nil)}
		set_menu_hint(i, 'Add a symbol')
		filemenu.append(i)

		i = Gtk::ImageMenuItem.new
		filemenu.append(i)

		name = 'Save'
		i = PetedMenuItemStock.new(title, name, Gtk::Stock::SAVE)
		i.signal_connect('activate') {|w| main_window.save_schematic(false)}
		set_menu_hint(i, 'Save the current file')
		filemenu.append(i)

		name = 'Save As...'
		i = PetedMenuItemStock.new(title, name, Gtk::Stock::SAVE_AS)
		i.signal_connect('activate') {|w| main_window.save_schematic(true)}
		set_menu_hint(i, 'Save the current file with a different name')
		filemenu.append(i)

		i = Gtk::ImageMenuItem.new
		filemenu.append(i)

		name = 'Close'
		i = PetedMenuItemStock.new(title, name, Gtk::Stock::CLOSE)
		i.signal_connect('activate') {|w| main_window.close_schematic}
		set_menu_hint(i, 'Close the current sheet')
		filemenu.append(i)

		name = 'Quit'
		i = PetedMenuItemStock.new(title, name, Gtk::Stock::QUIT)
		i.signal_connect('activate') {|w| main_window.destroy}
		set_menu_hint(i, 'Quit the program')
		filemenu.append(i)

		title = 'Edit'
		editmenu = Gtk::Menu.new
		editmenu.set_accel_group(group)
		editmenu.set_accel_path(MenuWindowName + '/' + title + '/')
		edit = Gtk::MenuItem.new(title)
		edit.submenu = editmenu
		self.append(edit)

		name = 'Undo'
		i = PetedMenuItemStock.new(title, name, Gtk::Stock::UNDO)
		i.sensitive = false
		i.signal_connect('activate') {|w| puts title + name + ' selected'}
		set_menu_hint(i, 'Undo the last action')
		editmenu.append(i)
	end
	
	def set_menu_hint(item, text)
		item.signal_connect('enter_notify_event') {|w, e| @bar.push(@id, text); false}
		item.signal_connect('leave_notify_event') {|w, e| @bar.pop(@id); false}
	end
end

class GridSelectBox < Gtk::ComboBoxEntry
	PAT = /^\s*\d{1,3}\s*$/
	def initialize(radio_button, grid, major)
		super()
		@major = major
		@adds = 0
		self.width_request = 80 # show 3 digits -- is there a way to specify width with num characters?
		self.tooltip_text = (@major ? 'Major grid' : 'Minor grid')
		default = grid[-1]
		@s = grid.uniq.sort
		default = @s.index(default)
		@s.map!{|el| el.to_s}
		@s.each{|g| self.append_text(g)}
		self.active = default
		self.child.signal_connect('populate-popup') {|w, menu|
			t = w.text
			item = Gtk::MenuItem.new
			item.show
			menu.append(item)
			item = Gtk::MenuItem.new('_Add to List')
			item.show
			item.signal_connect('activate') {|i|
				if i.sensitive? # should be true always -- see below
					self.prepend_text(t)
					@s << t
					@adds += 1
				end
			}
			menu.append(item)
			item.sensitive = (w.text_length > 0) && PAT.match(t) && !@s.include?(t)
			item = Gtk::MenuItem.new('_Remove Top')
			item.show
			item.signal_connect('activate') {|i|
				if @adds > 0
					self.remove_text(0)
					@s.pop
					@adds -= 1
				end
			}
			menu.append(item)
			item.sensitive = (@adds > 0)
		}
		self.child.signal_connect('activate') {|w|
			if @pda
				if PAT.match(w.text)
					i = w.text.to_i
					if i != (@major ? @pda.schem.major_grid : @pda.schem.minor_grid)
						if @major then @pda.schem.major_grid = i else @pda.schem.minor_grid = i end
						@pda.schem.active_grid = i if radio_button.active?
						@pda.redraw
					end
				else
					w.text = (@major ? @pda.schem.major_grid : @pda.schem.minor_grid).to_s
				end
			end
		}
		@changed_handler_id = self.signal_connect('changed') {|w|
			if @pda
				if w.active != -1
					i = w.active_text.to_i
					if i != (@major ? @pda.schem.major_grid : @pda.schem.minor_grid)
						if @major then @pda.schem.major_grid = i else @pda.schem.minor_grid = i end
						@pda.schem.active_grid = i if radio_button.active?
						@pda.redraw
					end
				end
			end
		}
	end

	def switch(pda)
		if self.sensitive = !!(@pda = pda) # maybe always sensitive
			self.signal_handler_block(@changed_handler_id)
			self.child.text = (@major ? @pda.schem.major_grid : @pda.schem.minor_grid).to_s
			self.signal_handler_unblock(@changed_handler_id)

		end
	end
end

class PetToolbar < Gtk::Toolbar
	def initialize(main_window, conf)
		super()
		@S1 = nil
		@open_button = Gtk::ToolButton.new(Gtk::Stock::OPEN)
		@open_button.signal_connect('clicked') {main_window.open_schematics(nil)}
		@open_button.tooltip_text = 'Open a file'

		@save_button = Gtk::ToolButton.new(Gtk::Stock::SAVE)
		@save_button.signal_connect('clicked') {main_window.save_schematic(false)}
		@save_button.tooltip_text = 'Save the current file'
		@save_button.sensitive = false

		@entry = Gtk::Entry.new
		@entry_item = Gtk::ToolItem.new
		@entry_item.add(@entry)
		@combo_box_mode = Gtk::ComboBox.new
		@combo_box_mode.signal_connect('changed') { |w, event| if @S1 then @S1.set_input_mode(w.active_text) end}
	Input_Mode.constants.each{|c| @combo_box_mode.append_text(c.to_s)}
 @combo_box_mode.active = Input_Mode::default
		@modes_item = Gtk::ToolItem.new
		@modes_item.add(@combo_box_mode)
@major_grid_select = Gtk::RadioToolButton.new(nil, Gtk::Stock::APPLY)
@major_grid_select.active = false
@major_grid_select.signal_connect('toggled') {|w|
	@S1.active_grid = (w.active? ? @S1.major_grid : @S1.minor_grid) if @S1
puts 'toggled major', @S1.active_grid if @S1
}
@minor_grid_select = Gtk::RadioToolButton.new(@major_grid_select, Gtk::Stock::APPLY)
@minor_grid_select.active = true
@minor_grid_select.signal_connect('toggled') {|w|
	@S1.active_grid = (w.active? ? @S1.minor_grid : @S1.major_grid) if @S1
}
	@combo_box_major_grid = GridSelectBox.new(@major_grid_select, Def::Major_Grid << conf.get_conf(Pet_Config::SCR_S)[:grid_size_major], true)
	@major_grid_item = Gtk::ToolItem.new
	@major_grid_item.add(@combo_box_major_grid)
	@combo_box_minor_grid = GridSelectBox.new(@minor_grid_select, Def::Minor_Grid << conf.get_conf(Pet_Config::SCR_S)[:grid_size_minor], false)
	@minor_grid_item = Gtk::ToolItem.new
	@minor_grid_item.add(@combo_box_minor_grid)
		self.insert(-1, @open_button)
		self.insert(-1, @save_button)
		self.insert(-1, @entry_item)
		self.insert(-1, @modes_item)
		self.insert(-1, @major_grid_select)
		self.insert(-1, @major_grid_item)
		self.insert(-1, @minor_grid_select)
		self.insert(-1, @minor_grid_item)
	end

	def switch(pda)
		if (@pda = pda)
			@S1 = @pda.schem
			@combo_box_major_grid.switch(pda)
			@combo_box_minor_grid.switch(pda)
			if @pda.schem.major_grid == @pda.schem.active_grid
				@major_grid_select.active = true
			else
				@minor_grid_select.active = true
			end
		else
			@S1 = nil
		end
		@save_button.sensitive = !!@S1
	end

	def init_grid(pda)
		pda.schem.major_grid = @combo_box_major_grid.child.text.to_i
		pda.schem.minor_grid = @combo_box_minor_grid.child.text.to_i
		pda.schem.active_grid =  (@major_grid_select.active? ? pda.schem.major_grid : pda.schem.minor_grid)
	end

end

class Pet_Canvas::PDA
	attr_accessor :hit, :ebdx , :ebdy
end

class Main_Window < Gtk::Window
	attr_reader :attr_box

	def push_msg(text)
		@statusbar2.push(0, text)
	end

	def pop_msg
		@statusbar2.pop(0)
	end

	def run_schematics_dialog(parent)
		dialog = Gtk::FileChooserDialog.new('Open Files', parent, Gtk::FileChooser::ACTION_OPEN, nil, # filesystem backend
			[Gtk::Stock::CANCEL, Gtk::Dialog::RESPONSE_CANCEL], [Gtk::Stock::OPEN, Gtk::Dialog::RESPONSE_ACCEPT])
		[['Schematics', '*.sch'], ['All Files', '*']].each{|el|
			filter = Gtk::FileFilter.new
			filter.name = el[0]
			filter.add_pattern(el[1])
			dialog.add_filter(filter)
		}
		dialog.select_multiple = true
		names = nil
		dialog.run do |response|
			if response == Gtk::Dialog::RESPONSE_ACCEPT
				names = dialog.filenames.dup
			end
		end
		dialog.destroy
		names
	end


	def run_schematic_save_dialog(parent, name)
		dialog = Gtk::FileChooserDialog.new('Save File', parent, Gtk::FileChooser::ACTION_SAVE, nil, # filesystem backend
			[Gtk::Stock::CANCEL, Gtk::Dialog::RESPONSE_CANCEL], [Gtk::Stock::SAVE, Gtk::Dialog::RESPONSE_ACCEPT])
		[['Schematics', '*.sch'], ['All Files', '*']].each{|el|
			filter = Gtk::FileFilter.new
			filter.name = el[0]
			filter.add_pattern(el[1])
			dialog.add_filter(filter)
		}
		#dialog.select_multiple = true
		#name = nil
		dialog.filename = name
		dialog.run do |response|
			if response == Gtk::Dialog::RESPONSE_ACCEPT
				name = dialog.filename
			else
				name = nil
			end
		end
		dialog.destroy
		name
	end





	def run_add_symbol_dialog(parent)
		dialog = Gtk::FileChooserDialog.new('Select Symbol', parent, Gtk::FileChooser::ACTION_OPEN, nil, # filesystem backend
			[Gtk::Stock::CANCEL, Gtk::Dialog::RESPONSE_CANCEL], [Gtk::Stock::OPEN, Gtk::Dialog::RESPONSE_ACCEPT])
		[['Symbols', '*.sym'], ['All Files', '*']].each{|el|
			filter = Gtk::FileFilter.new
			filter.name = el[0]
			filter.add_pattern(el[1])
			dialog.add_filter(filter)
		}
		dialog.select_multiple = false
		dialog.current_folder	= '/usr/share/gEDA/sym'
		name = nil
		dialog.run do |response|
			if response == Gtk::Dialog::RESPONSE_ACCEPT
				name = dialog.filename.dup
			end
		end
		dialog.destroy
		name
	end

	def initialize
		super
		@conf = Pet_Config::get_default_config()
		group = Gtk::AccelGroup.new
		@statusbar1 = Gtk::Statusbar.new
		@statusbar1.has_resize_grip = false
		@statusbar1.set_size_request(200, -1)
		menubar = PetedMenuBar.new(group, self, @statusbar1)
		@toolbar_top = PetToolbar.new(self, @conf)
		@vbox = Gtk::VBox.new(false, 5)
		@paned = Gtk::HPaned.new
		@vpaned = Gtk::VPaned.new
		@vbox.pack_start(menubar, false, false, 0)
		@vbox.pack_start(@toolbar_top, false, false, 0)
		@mode = Mode_Auto
		@selecting = false
		@hit = true
		signal_connect('destroy') { Gtk.main_quit }
		title = 'PetEd'
		set_size_request(600, 400)
		@notebook = Gtk::Notebook.new
		@attr_box = Pet_Attr::Attr_Win.new
		#@pda.schem.set_dialog_widget(self.attr_box)
		@notebook.append_page(@attr_box, Gtk::Label.new('Attr'))
		#conf.canvas = @pda
		@cnb = Pet_Conf_Ed::Config_Notebook.new(@conf)
		@notebook.append_page(@cnb, Gtk::Label.new('Conf'))
		@logview = Log::Log_View.new(Gtk::POLICY_ALWAYS, Gtk::POLICY_ALWAYS)
		@schematics = PetedNotebookWidget.new(@toolbar_top)
		@schematics.signal_connect('switch-page') do |w, page, page_num|
				child = w.get_nth_page(page_num)
				@toolbar_top.switch(child)
		end

		@vpaned.pack2(@logview, false, true)
		@vpaned.pack1(@schematics, true, true)
		@swin = Gtk::ScrolledWindow.new
		@swin.add_with_viewport(@notebook)
		@paned.pack1(@swin, false, true)
		@paned.pack2(@vpaned, true, true)
		@paned.position = 300
		@vbox.pack_start(@paned, true, true, 0)
		@status_box = Gtk::HBox.new(false, 5)
		@statusbar2 = Gtk::Statusbar.new
		@statusbar2.set_size_request(400, -1)
		@statusbar2.has_resize_grip = false
		@status_box.pack_start(@statusbar1, true, true, 0)
		@status_box.pack_end(@statusbar2, false, false, 0)
		@vbox.pack_start(@status_box, false, false, 0)
		add @vbox
		add_accel_group(group)
show_all
	end

  Unnamed = 'Unsaved Document'
	UNAME_PAT = Regexp.new(Unnamed + ' (\d+)')
	def open_schematics(names)
		names ||= run_schematics_dialog(self)
		return unless names
		names.each{|n|
			s = Pet::Schem.new
			unless n == ''
				if File.exist?(n)
					if s.ProcessInputFile(n) # method will log failure
						Log::puts("Read file #{n}")
					end
				else
					Log::puts("File #{n} does not exist -- using new sheet")
				end
			end
			s.filename = n
			pda = Pet_Canvas::PDA.new(s)
			s.pda = pda
			s.main_window = self
			pda.darea.can_focus = true
			pda.darea.add_events(Gdk::Event::BUTTON_PRESS_MASK | Gdk::Event::BUTTON_RELEASE_MASK | Gdk::Event::SCROLL_MASK |
				Gdk::Event::BUTTON1_MOTION_MASK | Gdk::Event::BUTTON2_MOTION_MASK | Gdk::Event::POINTER_MOTION_HINT_MASK|
				Gdk::Event::POINTER_MOTION_MASK | Gdk::Event::KEY_PRESS_MASK | Gdk::Event::ENTER_NOTIFY_MASK | Gdk::Event::LEAVE_NOTIFY_MASK)
			pda.darea.signal_connect('motion-notify-event')  {|w, e| distribute_events(pda, e)}
			pda.darea.signal_connect('button_press_event')   {|w, e| distribute_events(pda, e)}
			pda.darea.signal_connect('button_release_event') {|w, e| distribute_events(pda, e)}
			pda.darea.signal_connect('key_press_event')      {|w, e| distribute_events(pda, e)}
			pda.darea.signal_connect('enter-notify-event')   {|w, e| pda.window.cursor = pda.cursor; false}
			pda.darea.signal_connect('leave-notify-event')   {|w, e| pda.window.cursor = nil; false}
			pda.darea.can_focus = true
			@conf.canvas = pda
			@toolbar_top.init_grid(pda)
			pda.show_all
			if n.empty?
			  busy = [0]
				i = 0
				while (page = @schematics.get_nth_page(i))
					t = @schematics.get_labeltext(page)
					if m = UNAME_PAT.match(t)
					  busy << m[1].to_i
					end
					i += 1
				end
				busy.sort!
				busy << 1e4.to_i
				i = 0
				until i < busy[i]; i += 1; end
				n = "Unsaved Document #{i}"
			end
			@schematics.append_page(pda, n)
			@schematics.page = -1
			pda.schem.set_dialog_widget(self.attr_box)
			pda.darea.grab_focus
		}
	end

	def save_schematic(save_as)
		if (i = @schematics.page) != -1
			pda = @schematics.get_nth_page(i)
			name = pda.schem.filename
			name = run_schematic_save_dialog(self, name) if save_as or name.empty?
			if name and !name.empty?
 				pda.schem.filename = name
			  pda.schem.write(name)
				@schematics.update_label(pda, name)
			end
		end
	end

	def close_schematic
		@schematics.remove_page(@schematics.page)
		@toolbar_top.switch( @schematics.get_nth_page(@schematics.page))
	end

	def add_symbol(name)
		name = run_add_symbol_dialog(self)
			i = @schematics.page
			pda = @schematics.get_nth_page(i)
		pda.schem.ProcessSymFile(name)
	end

	def close_all
		puts 'close_all'
	end

	def show_attributes(show)
		puts 'show_attributes == ', show
	end

	def distribute_events(pda, event)
		if event.event_type == Gdk::Event::KEY_PRESS
			px, py = pda.get_user_coordinates(*pda.pointer)
		else
			px, py = pda.get_user_coordinates(event.x, event.y)
			if (event.event_type == Gdk::Event::MOTION_NOTIFY)
				if (event.state & Gdk::Window::BUTTON1_MASK) != 0
					unless pda.hit
						pda.draw_select_rect(pda.ebdx , pda.ebdy, event.x, event.y)
						#return
					end
				elsif (event.state & Gdk::Window::BUTTON2_MASK) != 0
					if pda.hit
						pda.pan(pda.ebdx - event.x, pda.ebdy - event.y)
						pda.ebdx , pda.ebdy = event.x, event.y
					else
						pda.draw_select_rect(pda.ebdx , pda.ebdy, event.x, event.y)
					end
					return
				end
			elsif event.event_type == Gdk::Event::BUTTON_PRESS

				pda.darea.grab_focus
				pda.ebdx , pda.ebdy = event.x, event.y
				pda.hit = pda.schem.hoovering?
			elsif event.event_type == Gdk::Event::BUTTON_RELEASE
				if not pda.hit
					if event.button == 2
						pda.zoom_into_select_rect
						return
					else
						pda.draw_select_rect(0, 0, 0, 0)
					end
				end
			end
		end
		unless ((event.event_type == Gdk::Event::MOTION_NOTIFY) and ((event.state & Gdk::Window::BUTTON2_MASK) != 0)) or
					 (((event.event_type == Gdk::Event::BUTTON_PRESS) or (event.event_type == Gdk::Event::BUTTON_RELEASE)) and (event.button == 2))
			boxlist = Array.new
			pda.schem.process_event(boxlist, event, px, py, 100) # grid==100 for now
			pda.update_canvas(boxlist)
		end
	end

end # Peted

#Gtk.init # Since Ruby-GNOME2-0.15.0, this is included in gtk2.rb, so we don't need call this

options = Hash.new
optparse = OptionParser.new do |opts|
	opts.banner = "Usage: peted.rb [options] file1 file2 ..."
	options[:verbose] = false
	opts.on('-v', '--verbose', 'Output more information') do
		 options[:verbose] = true
	end
	options[:logfile] = nil
	opts.on( '-l', '--logfile FILE', 'Write log to FILE' ) do|file|
		options[:logfile] = file
	end

	opts.on('-h', '--help', 'Display this screen') do
		 puts opts
		 exit
	end
end
optparse.parse!

main_window = Main_Window.new
main_window.open_schematics(ARGV)

Gtk.main

end # module PetEd

